import logging

_default_logger = logging.getLogger('videoflow')

def set_verbosity(verbosity : str = "WARNING"):
    verbosity_map = {
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'WARNING': logging.WARNING,
        'ERROR': logging.ERROR,
        'CRITICAL': logging.CRITICAL
    }

    if verbosity not in verbosity_map:
        raise ValueError("Logging verbosity is not one of 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL")
    
    _default_logger.setLevel(verbosity_map[verbosity])
    
def debug(message, *args, **kwargs):
    _default_logger.debug(message, *args, **kwargs)

def error(message, *args, **kwargs):
    _default_logger.error(message, *args, **kwargs)

def warning(message, *args, **kwargs):
    _default_logger.warning(message, *args, **kwargs)

def info(message, *args, **kwargs):
    _default_logger.info(message, *args, **kwargs)

def critical(message, *args, **kwargs):
    _default_logger.critical(message, *args, **kwargs)


import subprocess

def get_number_of_gpus() -> int:
    '''
    Returns the number of gpus in the system
    '''
    try:
        n = str(subprocess.check_output(["nvidia-smi", "-L"])).count('UUID')
        return n
    except FileNotFoundError:
        return 0
    


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import time

from ..core.node import ProducerNode

class IntProducer(ProducerNode):
    '''
    Each time the ``next`` method is called, produces the next
    integer in the sequence between ``start_value`` and ``end_value``.
    
    If ``wait_time_in_seconds`` is greater than zero, it sleeps for 
    the specified amount of seconds each time ``next()`` is called
    '''
    def __init__(self, start_value : int = 0, end_value : int = None, 
                wait_time_in_seconds : float = 0):
        self._start_value = start_value
        self._end_value = end_value
        self._wait_time_in_seconds = wait_time_in_seconds
        self._current_value = self._start_value
        super(IntProducer, self).__init__()
        
    def next(self):
        '''
        - Returns:
            - next: an integer
        '''
        if self._end_value is not None and self._current_value > self._end_value:
            raise StopIteration()
        to_return = self._current_value
        self._current_value += 1
        time.sleep(self._wait_time_in_seconds)
        return to_return



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .node import Node, ProducerNode, ConsumerNode, ProcessorNode
from .task import Task, ProducerTask, ProcessorTask, ConsumerTask, STOP_SIGNAL
from ..environments.queues import RealtimeQueueExecutionEnvironment, BatchprocessingQueueExecutionEnvironment
from ..core import logger

BATCH = 'batch'
REALTIME = 'realtime'
flow_types = [BATCH, REALTIME]

def _has_cycle_util(v : Node, visited, rec):
    visited[v] = True
    rec[v] = True
    
    for child in v.children:
        if not child in visited:
            visited[child] = False
        if visited[child] == False:
            if _has_cycle_util(child, visited, rec):
                return True
        elif rec[child] == True:
            return True
    
    rec[v] = False
    return False

def has_cycle(producers):
    '''
    Used to detect if the graph is not acyclical.  Returns true if it \
    finds a cycle in the graph.  It begins exploring the graph from producers down \
    all the way to consumers.
    '''
    visited = {}
    rec = {}
    for v in producers:
        visited[v] = False
        rec[v] = False
    
    for v in producers:
        if visited[v] == False:
            if _has_cycle_util(v, visited, rec):
                return True
    return False
    
def _topological_sort_util(v : Node, visited, stack):
    if not v in visited:
        visited[v] = False
    visited[v] = True
    for child in v.children:
        if not child in visited or visited[child] == False:
            _topological_sort_util(child, visited, stack)
    stack.insert(0, v)

def topological_sort(producers):
    '''
    Creates a topological sort of the computation graph.

    - Arguments:
        - producers: a list of producer nodes, that is, nodes with no parents.
    
    - Returns:
        - stack: a list of nodes in topological order.  If \
            a *node A* appears before a *node B* on the list, it means \
            that *node A* does not depend on *node B* output
    '''
    visited = {}
    for v in producers:
        visited[v] = False
    stack = []

    for v in producers:
        if visited[v] == False:
            _topological_sort_util(v, visited, stack)
    
    return stack

class Flow:
    '''
    Represents a linear flow of data from one task to another.\
    Note that a flow is created from a **directed acyclic graph** of producer, processor \
    and consumer nodes, but the flow itself is **linear**, because it is an optimized \
    `topological sort` of the directed acyclic graph.

    - Arguments:
        - producers: a list of producer nodes of type ``videoflow.core.node.ProducerNode``.
        - consumers: a list of consumer nodes of type ``videoflow.core.node.ConsumerNode``.
    '''
    def __init__(self, producers, consumers, flow_type = REALTIME):
        if len(producers) != 1:
            raise AttributeError('Only support flows with 1 producer for now.')
        self._producers = producers
        self._consumers = consumers
        self._tasks = None
        self._producer_tasks = []
        
        if flow_type == REALTIME:
            self._execution_environment = RealtimeQueueExecutionEnvironment()
        elif flow_type == BATCH:
            self._execution_environment = BatchprocessingQueueExecutionEnvironment()
        else:
            raise ValueError('flow_type must be one of {}'.format(','.join(flow_types)))

    def run(self):
        '''
        Simple documentation: It starts the flow. 

        More complex documentation: 
        
        1. It creates a topological sort of the nodes in the \
            computation graph, and wraps each node around a ``videoflow.core.task.Task``
        2. It passes the tasks to the environment, which allocates them and creates the \
            channels that will be used for communication between tasks. Tasks themselves \
            do not know where this channels are, but the environment assigns a messenger \
            to each task that knows how to communicate in those channels.
        '''

        #1. Build a topological sort of the graph.
        if has_cycle(self._producers):
            logger.error('Cycle detected in computation graph. Exiting now...')
            raise ValueError('Cycle found in graph')

        #2. TODO: CHeck that all nodes in the graph are
        # descendants of a producer
        #3. TODO: Check that all producers' results are
        #being read by a consumer.

        tsort = topological_sort(self._producers)
        #2. TODO: OPtimize graph in the following ways:   
        # a) Tasks do not need to pass down to children
        # all of the outputs of parents.  Hence, at a given
        # level of the topological sort, have the list of 
        # inputs from parents that are not needed below that 
        # level

        # b) Not all the processors have to write to a pub/sub channel
        # If their output is only needed by the next preprocessor and non one
        # else below in the graph, then I can string subsequent preprocessors together
        # a big preprocessor
        
        #3. Create the tasks and the input/outputs
        # for them
        tasks = []
        for i in range(len(tsort)):
            node = tsort[i]
            
            if isinstance(node, ProducerNode):
                task = ProducerTask(node, i)
                self._producer_tasks.append(task)
            elif isinstance(node, ProcessorNode):
                task = ProcessorTask(
                    node, 
                    i,
                    i - 1
                )
            elif isinstance(node, ConsumerNode):
                task = ConsumerTask(
                    node,
                    i,
                    i - 1,
                    i < (len(tsort) - 1)
                )
            else:
                raise ValueError('node is not of one of the valid types')
            tasks.append(task)
        
        # 4. Put each task to run in the place where the processor it
        # contains inside runs.
        logger.info('Started task allocation for {} tasks'.format(len(tasks)))
        self._execution_environment.allocate_and_run_tasks(tasks)
        logger.info('Started running flow.')
    
    def join(self):
        '''
        Will make the process that calls this method block until the flow finishes
        running naturally
        '''
        self._execution_environment.join_task_processes()


    def stop(self):
        '''
        Stops the flow.  Makes the execution environment send a flow termination signal.
        '''
        self._execution_environment.signal_flow_termination()



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .node import Node, ProducerNode, ProcessorNode, ConsumerNode

STOP_SIGNAL = "alalsl;'sdlfj2389jdalskmghsaiaw98y8asdf;askljoa8y;dsf;lkasdb"

class Task:
    '''
    A ``Task`` is a wrapper around a ``videoflow.core.node.Node`` that \
        is able to interact with the execution environment through a messenger. \
        Nodes receive input and/or produce output, but tasks are the ones \
        that run in infinite loops, receiving inputs from the environment and passing them to the \
        computation node, and taking outputs from the computation node and passing \
        them to the environment.
    '''
    def __init__(self, computation_node : Node, task_id : int, parent_task_id : int = None):
        self._messenger = None
        self._computation_node = computation_node
        self._task_id = task_id
        self._parent_task_id = parent_task_id
    
    @property
    def id(self):
        '''
        Returns an integer as id.
        '''
        return self._task_id
    
    @property
    def parent_id(self):
        '''
        Returns the id of the parent task.  Id of parent task is lower than id of current task.
        '''
        return self._parent_task_id
    
    @property
    def computation_node(self):
        '''
        Returns the current computation node
        '''
        return self._computation_node

    def set_messenger(self, messenger):
        '''
        Used by environment to set the messenger that this task will use to interact with other 
        tasks
        '''
        self._messenger = messenger

    def _assert_messenger(self):
        assert self._messenger is not None, 'Task cannot run if messenger has not been set.'

    def _run(self):
        raise NotImplementedError('Sublcass needs to implement _run')

    def run(self):
        '''
        Starts the task in an infinite loop.  If this method is called and the \
            ``set_messenger()`` method has not been called yet, an assertion error \
            will happen.
        '''
        self._assert_messenger()
        self._computation_node.open()
        self._run()
        self._computation_node.close()

class ProducerTask(Task):
    '''
    It runs forever calling the ``next()`` method in the producer node. \
    At each iteration it checks for a termination signal, and if so it \
    sends a termination message to its child task and breaks the infinite loop.
    '''
    def __init__(self, producer : ProducerNode, task_id : int):
        self._producer = producer
        super(ProducerTask, self).__init__(producer, task_id)
    
    def _run(self):
        while True:
            try:
                a = self._producer.next()
                self._messenger.publish_message(a)
            except StopIteration:
                break
            if self._messenger.check_for_termination():
                break
        self._messenger.publish_termination_message(STOP_SIGNAL)

class ProcessorTask(Task):
    '''
    It runs forever, first blocking until it receives a message from parent nodes through \
    the messenger.  Then it passes it to the processor node and when it gets back the output \
    it uses the messenger to publish it down the flow. If among the inputs it received from \
    a parent it receives a termination message, it passes termination message down the flow \
    and breaks from infinite loop.
    '''
    def __init__(self, processor : ProcessorNode, task_id : int, parent_task_id : int):
        self._processor = processor
        
        super(ProcessorTask, self).__init__(processor, task_id, parent_task_id)    
    
    @property
    def device_type(self):
        return self._processor.device_type
    
    def change_device(self, device_type : str):
        self._processor.change_device(device_type)
    
    def _run(self):
        while True:
            inputs = self._messenger.receive_message()
            stop_signal_received = any([isinstance(a, str) and a == STOP_SIGNAL for a in inputs])
            if stop_signal_received:
                self._messenger.publish_termination_message(STOP_SIGNAL)
                break

            #3. Pass inputs needed to processor
            output = self._processor.process(*inputs)
            self._messenger.publish_message(output)
        
class ConsumerTask(Task):
    '''
    It runs forever, blocking until it receives a message from parent nodes through the messenger.
    It consumes the message and does not publish anything back down the pipe.

    If a consumer task has tasks after it in the topological sort, it does not mean that
    those tasks expect any input from the consumer task. It simply means that the consumer
    task is a passthrough of messages. 
    '''
    def __init__(self, consumer : ConsumerNode, task_id : int, parent_task_id : int,
                has_children_task : bool):
        self._consumer = consumer
        self._has_children_task = has_children_task
        super(ConsumerTask, self).__init__(consumer, task_id, parent_task_id)
    
    def _run(self):
        while True:
            inputs = self._messenger.receive_message()
            stop_signal_received = any([isinstance(a, str) and a == STOP_SIGNAL for a in inputs])
            if stop_signal_received:
                # No need to pass through stop signal to children.
                # If children need to stop, they will receive it from
                # someone else, so the message that I am passing through
                # might be the one carrying it.
                if self._has_children_task:
                    self._messenger.passthrough_termination_message()
                break

            if self._has_children_task:
                self._messenger.passthrough_message()
            self._consumer.consume(*inputs)
    


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

class Processor:
    def process(self, item):
        raise NotImplementedError('Subclass must implement method')



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

def parse_label_map(path_to_labels: str):
    '''
    - Arguments:
        - path_to_labels (str): path to pbtx file

    - Returns:
        - dict of form { id(int) : label(str)}
    '''
    with open(path_to_labels, "r") as f:
        text = f.read()
    entry_pairs = []
    a = text.find('item')
    while a != -1:
        b = text.find('id:', a)
        b1 = text.find('\n', b)
        index = int(text[b + len("id:"): b1])
        c = text.find('display_name:', a)
        c1 = max(text.find("\'", c), text.find('\"', c))
        c2 = max(text.find("\'", c1 + 1), text.find('\"', c1 + 1))
        klass_name = text[c1 + 1: c2]
        entry_pairs.append((index, klass_name))
        a = text.find('item', c)
    return dict(entry_pairs)



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
from filterpy.kalman import KalmanFilter
from sklearn.utils.linear_assignment_ import linear_assignment
import math

from ...core.node import OneTaskProcessorNode

class BoundingBoxTracker(OneTaskProcessorNode):
    '''
    Tracks bounding boxes from one frame to another.
    It keeps an internal state representation that allows
    it to track across frames.
    '''
    def _track(self, dets : np.array) -> np.array:
        '''
        - Arguments: 
            - dets: np.array of shape (nb_boxes, 6) \
                Specifically (nb_boxes, [ymin, xmin, ymax, xmax, class_index, score])
        '''
        raise NotImplementedError("Subclass must implement _track method")
    
    def process(self, dets : np.array) -> np.array:
        '''
        - Arguments: 
            - dets: np.array of shape (nb_boxes, 6) \
                Specifically (nb_boxes, [ymin, xmin, ymax, xmax, class_index, score])
        - Returns:
            - tracks: np.array of shape (nb_boxes, 5) \
                Specifically (nb_boxes, [ymin, xmin, ymax, xmax, track_id])
        '''
        return self._track(dets)

def eucl(bb_test, bb_gt):
    '''
    Computes the euclidean distance between two boxes
    in the form  y1, x1, y2, x2]
    '''
    center_1 = [(bb_test[0] + bb_test[2]) / 2.0, (bb_test[1] + bb_test[3]) / 2.0]
    center_2 = [(bb_gt[0] + bb_gt[2]) / 2.0, (bb_gt[1] + bb_gt[3]) / 2.0]
    eucl = math.sqrt((center_1[0] - center_2[0])*(center_1[0] - center_2[0]) + (center_1[1] - center_2[1])*(center_1[1] - center_2[1]))
    return -eucl

def iou(bb_test, bb_gt):
    """
    Computes IUO between two bboxes in the form [ y1, x1, y2, x2]
    IOU is the intersection of areas.
    """
    yy1 = np.maximum(bb_test[0], bb_gt[0])
    xx1 = np.maximum(bb_test[1], bb_gt[1])
    yy2 = np.minimum(bb_test[2], bb_gt[2])
    xx2 = np.minimum(bb_test[3], bb_gt[3])
    w = np.maximum(0., xx2 - xx1)
    h = np.maximum(0., yy2 - yy1)
    wh = w * h
    o = wh / ((bb_test[2] - bb_test[0]) * (bb_test[3] - bb_test[1])
              + (bb_gt[2] - bb_gt[0]) * (bb_gt[3] - bb_gt[1]) - wh)
    return(o)

def convert_bbox_to_z(bbox):
    """
    Takes a bounding box in the form [ y1, x1, y2, x2] and returns z in the form
    [x, y, s, r] where x, y is the centre of the box and s is the scale/area and r is
    the aspect ratio
    """
    w = bbox[1] - bbox[3]
    h = bbox[0] - bbox[2]
    x = bbox[1] + w/2.
    y = bbox[0] + h/2.
    s = w * h    #scale is just area
    r = w / float(h)
    return np.array([x, y, s, r]).reshape((4, 1))

def convert_x_to_bbox(x, score=None):
    """
    Takes a bounding box in the form [x, y, s, r] and returns it in the form
    [y1, x1, y2, x2] where x1, y1 is the top left and x2, y2 is the bottom right
    """
    w = np.sqrt(x[2]*x[3])
    h = x[2]/w
    if(score==None):
        return np.array([ x[1] - h/2., x[0] - w/2.,  x[1] + h/2., x[0] + w/2.,]).reshape((1, 4))
    else:
        return np.array([ x[1] - h/2.,x[0] - w/2.,  x[1] + h/2., x[0] + w/2., score]).reshape((1, 5))

def associate_detections_to_trackers(detections, trackers, metric_function, iou_threshold = 0.1):
    """
    Assigns detections to tracked object (both represented as bounding boxes)
    Returns 3 lists of matches, unmatched_detections and unmatched_trackers
    """
    distance_threshold = 500

    if(len(trackers) == 0):
        return np.empty((0, 2), dtype = int), np.arange(len(detections)), np.empty((0, 5), dtype = int)
    iou_matrix = np.zeros((len(detections), len(trackers)), dtype=np.float32)

    for d, det in enumerate(detections):
        for t, trk in enumerate(trackers):
            iou_matrix[d, t] = metric_function(det, trk)
    matched_indices = linear_assignment(-iou_matrix)

    unmatched_detections = []
    for d,det in enumerate(detections):
        if(d not in matched_indices[:,0]):
            unmatched_detections.append(d)
    unmatched_trackers = []
    for t, trk in enumerate(trackers):
        if(t not in matched_indices[:,1]):
            unmatched_trackers.append(t)

    #filter out matched with low IOU
    matches = []
    for m in matched_indices:
        if(iou_matrix[m[0], m[1]] < iou_threshold):
        #if(iou_matrix[m[0], m[1]] > distance_threshold):
            unmatched_detections.append(m[0])
            unmatched_trackers.append(m[1])
        else:
            matches.append(m.reshape(1, 2))
    if(len(matches)==0):
        matches = np.empty((0, 2), dtype = int)
    else:
        matches = np.concatenate(matches, axis = 0)

    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)

class KalmanBoxTracker(object):
    """
    This class represents the internel state of individual tracked objects observed as bbox.
    """
    count = 0
    def __init__(self,bbox):
        """
        Initialises a tracker using initial bounding box.
        """
        #define constant velocity model
        self.kf = KalmanFilter(dim_x=7, dim_z=4)
        self.kf.F = np.array([[1,0,0,0,1,0,0],[0,1,0,0,0,1,0],[0,0,1,0,0,0,1],[0,0,0,1,0,0,0],[0,0,0,0,1,0,0],[0,0,0,0,0,1,0],[0,0,0,0,0,0,1]])
        self.kf.H = np.array([[1,0,0,0,0,0,0],[0,1,0,0,0,0,0],[0,0,1,0,0,0,0],[0,0,0,1,0,0,0]])

        self.kf.R[2:,2:] *= 10.
        self.kf.P[4:,4:] *= 1000. #give high uncertainty to the unobservable initial velocities
        self.kf.P *= 10.
        self.kf.Q[-1,-1] *= 0.01
        self.kf.Q[4:,4:] *= 0.01

        self.kf.x[:4] = convert_bbox_to_z(bbox)
        self.time_since_update = 0
        self.id = KalmanBoxTracker.count
        KalmanBoxTracker.count += 1
        self.history = []
        self.hits = 0
        self.hit_streak = 0
        self.age = 0

    def update(self, bbox):
        """
        Updates the state vector with observed bbox.
        """
        self.time_since_update = 0
        self.history = []
        self.hits += 1
        self.hit_streak += 1
        self.kf.update(convert_bbox_to_z(bbox))

    def predict(self):
        """
        Advances the state vector and returns the predicted bounding box estimate.
        """
        if((self.kf.x[6] + self.kf.x[2]) <= 0):
            self.kf.x[6] *= 0.0
        self.kf.predict()
        self.age += 1
        if(self.time_since_update > 0):
            self.hit_streak = 0
        self.time_since_update += 1
        self.history.append(convert_x_to_bbox(self.kf.x))
        
        return self.history[-1]
    
    def get_state(self):
        """
        Returns the current bounding box estimate.
        """
        return convert_x_to_bbox(self.kf.x)


def metric_factory(metric_type):
    if metric_type == "iou":
        return iou
    elif metric_type == "euclidean":
        return eucl
    else:
        raise ValueError('Invalid metric type')

class KalmanFilterBoundingBoxTracker(BoundingBoxTracker):
    
    def __init__(self, max_age = 7, min_hits = 3, metric_function_type = 'iou'):
        self.max_age = max_age
        self.min_hits = min_hits
        self.trackers = []
        self.frame_count = 0
        self.metric_function_type = metric_function_type
        self.previous_fid = -1
        self.metric_function = metric_factory(metric_function_type)
        super(KalmanFilterBoundingBoxTracker, self).__init__()

    def _track(self, dets, fid = None):
        """
        Requires: this method must be called once for each frame even with empty detections.

        - Arguments:
            - dets: a numpy array of detections in the format [[ymin,xmin,ymax,xmax,score],[ymin,xmin,ymax,xmax,score],...]
                
        - Returns:
            - A similar array, where the last column is the object or track id.  The number of objects returned may differ from the number of detections provided.
        """
        if fid is None:
            fid = self.previous_fid + 1
        
        self.frame_count += 1
        #get predicted locations from existing trackers.
        trks = np.zeros((len(self.trackers), 5))
        to_del = []
        ret = []
        for t, trk in enumerate(trks):
            pos = self.trackers[t].predict()[0]
            trk[:] = [pos[0], pos[1], pos[2], pos[3], 0]
            if(np.any(np.isnan(pos))):
                to_del.append(t)
        trks = np.ma.compress_rows(np.ma.masked_invalid(trks))
        for t in reversed(to_del):
            self.trackers.pop(t)
        
        if self.metric_function_type == 'probability':
            index = int(fid - self.previous_fid)
            if index <= 9 and index > 0:
                tm = self.tm_powers[index]
            else: 
                tm = self.tm_powers[9]
            mf = probability_factory(tm, self.square_dims, self.frame_shape)
            matched, unmatched_dets, unmatched_trks = associate_detections_to_trackers(dets, trks, mf, iou_threshold = 0.00001)    
        else:
            matched, unmatched_dets, unmatched_trks = associate_detections_to_trackers(dets, trks, self.metric_function)

        #update matched trackers with assigned detections
        for t, trk in enumerate(self.trackers):
            if(t not in unmatched_trks):
                d = matched[np.where(matched[:,1] == t)[0], 0] 
                trk.update(dets[d,:][0])

        #create and initialise new trackers for unmatched detections
        for i in unmatched_dets:
            trk = KalmanBoxTracker(dets[i,:]) 
            self.trackers.append(trk)
        i = len(self.trackers)
        for trk in reversed(self.trackers):
            d = trk.get_state()[0]
            
            if((trk.time_since_update < 1) and (trk.hit_streak >= self.min_hits or self.frame_count <= self.min_hits)):
                ret.append(np.concatenate((d,[trk.id + 1])).reshape(1, -1)) # +1 as MOT benchmark requires positive
            i -= 1
        
            #remove dead tracklet
            if(trk.time_since_update > self.max_age):
                self.trackers.pop(i)
        if(len(ret) > 0):
            return np.concatenate(ret)
    
        self.previous_fid = fid
        return np.empty((0, 5))
        


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from ..core.node import ConsumerNode

class CommandlineConsumer(ConsumerNode):
    '''
    Writes the input received to the command line.
    
    - Arguments:
        - sep: separator to use between tokens.
        - end: end of line character
    '''
    def __init__(self, sep = ' ', end = '\n'):

        self._end = end
        self._sep = sep
        super(CommandlineConsumer, self).__init__()
    
    def consume(self, item):
        '''
        Prints `item` to the command line, adding an end of line character after it.
        - Arguments:
            - item: It can be anything that can be printed with the ``print()`` function
        '''
        print(item, sep = self._sep, end = self._end)



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import cv2

from ..core.node import ProducerNode

class VideofileReader(ProducerNode):
    '''
    Opens a video capture object and returns subsequent frames
    from the video each time ``next`` is called
    '''
    def __init__(self, video_file : str, nb_frames = -1):
        '''
        - Arguments:
            - video_file: path to video file
            - nb_frames: number of frames to process. -1 means all of them
        '''
        self._video_file = video_file
        self._video = None
        self._nb_frames = nb_frames
        self._frame_count = 0
        super(VideofileReader, self).__init__()
    
    def open(self):
        '''
        Opens the video stream
        '''
        if self._video is None:
            self._video = cv2.VideoCapture(self._video_file)

    def close(self):
        '''
        Releases the video stream object
        '''
        self._video.release()

    def next(self):
        '''
        - Returns:
            - frame: np.array of shape (h, w, 3)
        
        - Raises:
            - StopIteration: after it finishes reading the videofile \
                or when it reaches the specified number of frames to \
                process.
        '''
        
        if self._video.isOpened():
            success, frame = self._video.read()
            self._frame_count += 1
            if not success or self._frame_count == self._nb_frames:
                raise StopIteration()
            else:
                return frame
        else:
            raise StopIteration()



'''
Collection of object detection processors
'''
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
import tensorflow as tf

from ...core.node import ProcessorNode, CPU, GPU
from ...utils.tensorflow import TensorflowModel

class ObjectDetector(ProcessorNode):
    '''
    Abstract class that defines the interface of object detectors
    '''
    def _detect(self, im : np.array) -> np.array:
        '''
        - Arguments:
            - im (np.array): (h, w, 3)
        
        - Returns:
            - dets: np.array of shape (nb_boxes, 6) \
              Specifically (nb_boxes, [ymin, xmin, ymax, xmax, class_index, score])
        '''
        raise NotImplementedError('Subclass must implement it')
    
    def process(self, im : np.array) -> np.array:
        '''
        - Arguments:
            - im (np.array): (h, w, 3)
        
        - Returns:
            - dets: np.array of shape (nb_boxes, 6) \
                Specifically (nb_boxes, [ymin, xmin, ymax, xmax, class_index, score])
                The box coordinates are returned unnormalized (values NOT between 0 and 1, \
                but using the original dimension of the image)
        '''
        return self._detect(im)

class TensorflowObjectDetector(ObjectDetector):
    '''
    Finds object detections by running a Tensorflow model
    on an image.
    '''
    def __init__(self, path_to_pb_file,
                num_classes, 
                min_score_threshold = 0.5,
                nb_tasks = 1,
                device_type = CPU):
        '''
        Initializes the tensorflow model.  

        - Arguments:
            - path_to_pb_file (str): Path where model pb file is \
            It expects the model to have the following input tensors: `image_tensor:0`, and \
            the following output tensors: `detection_boxes:0`, `detection_scores:0`, \
            `detection_classes:0`, and `num_detections:0`
            - num_classes (int): number of classes that the detector can recognize
            - min_score_threshold (float): detection will filter out entries with score below threshold score
        '''
        self._tensorflow_model = None
        self._num_classes = num_classes
        self._path_to_pb_file = path_to_pb_file
        self._min_score_threshold = min_score_threshold
        super(TensorflowObjectDetector, self).__init__(nb_tasks = nb_tasks, device_type = device_type)
    
    def open(self):
        '''
        Creates session with tensorflow model
        '''
        if self.device_type == CPU:
            device_id = 'cpu'
        elif self.device_type == GPU:
            device_id = 'gpu'
        else:
            device_id = 'cpu'
        
        self._tensorflow_model = TensorflowModel(
            self._path_to_pb_file,
            ["image_tensor:0"],
            ["detection_boxes:0", "detection_scores:0", "detection_classes:0", "num_detections:0"],
            device_id = device_id
        )
    
    def close(self):
        '''
        Closes tensorflow model session.
        '''
        self._tensorflow_model._close_session()

    def _detect(self, im : np.array) -> np.array:
        '''
        - Arguments:
            - im (np.array): (h, w, 3)
        
        - Returns:
            - dets: np.array of shape (nb_boxes, 6) \
                Specifically (nb_boxes, [ymin, xmin, ymax, xmax, class_index, score])
        '''
        h, w, _ = im.shape
        im_expanded = np.expand_dims(im, axis = 0)
        boxes, scores, classes, num = self._tensorflow_model.run_on_input(im_expanded)
        boxes, scores, classes = np.squeeze(boxes, axis = 0), np.squeeze(scores, axis = 0), np.squeeze(classes, axis = 0)
        
        # boxes denormalization
        boxes[:,[0, 2]] = boxes[:,[0, 2]] * h
        boxes[:,[1, 3]] = boxes[:,[1, 3]] * w

        indexes = np.where(scores > self._min_score_threshold)[0]
        boxes, scores, classes = boxes[indexes], scores[indexes], classes[indexes]
        scores, classes = np.expand_dims(scores, axis = 1), np.expand_dims(classes, axis = 1)
        return np.concatenate((boxes, classes, scores), axis = 1)



from .task import Task, ConsumerTask, ProcessorTask, ProducerTask
from .node import Node, ConsumerNode, ProcessorNode, ProducerNode, ExternalProcessorNode, FunctionProcessorNode
from .flow import Flow


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from ..core.node import ProcessorNode

class IdentityProcessor(ProcessorNode):
    '''
    IdentityProcessor implements the identity
    function: it returns the same value that it received
    as input.
    '''
    def process(self, inp):
        return inp

class JoinerProcessor(ProcessorNode):
    '''
    Takes all the parameters received in the ``process`` method
    and makes them a tuple of items.
    '''
    def process(self, *inp):
        return tuple(inp)



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .processor import Processor

GPU = 'gpu'
CPU = 'cpu'
device_types = [CPU, GPU]

class Node:
    '''
    Represents a computational node in the graph. It is also a callable object. \
        It can be call with the list of parents on which it depends.
    '''
    def __init__(self):
        self._parents = None
        self._children = set()
    
    def __repr__(self):
        return self.__class__.__name__
    
    def __eq__(self, other):
        return self is other
    
    def __hash__(self):
        return id(self)
    
    def open(self):
        '''
        This method is called by the task runner before doing any consuming,
        processing or producing.  Should be used to open any resources
        that will be needed during the life of the task, such as opening files,
        tensorflow sessions, etc.
        '''
        pass
    
    def close(self):
        '''
        This method is called by the task running after finishing doing all
        consuming, processing or producing because of and end signal receival.
        Should be used to close any resources
        that were opened by the open() method, such as files,
        tensorflow sessions, etc.
        '''
        pass
    
    @property
    def id(self):
        '''
        The id of the node.  In this case the id of the node is produced by calling
        ``id(self)``.
        '''
        return self.__hash__()
    
    def __call__(self, *parents):
        if self._parents is not None:
            raise RuntimeError('This method has already been called. It can only be called once.')
        self._parents = list()
        for parent in parents:
            assert isinstance(parent, Node) and not isinstance(parent, Leaf), '%s is not a non-leaf node' % str(parent)
            self._parents.append(parent)
            parent.add_child(self)
        return self
        
    def add_child(self, child):
        '''
        Adds child to the set of childs that depend on it.
        '''
        self._children.add(child)
    
    @property
    def parents(self):
        '''
        Returns a list with the parent nodes
        '''
        if self._parents is None:
            return None
        return list(self._parents)
    
    @property
    def children(self):
        '''
        Returns a set of the child nodes
        '''
        return set(self._children)

class Leaf(Node):
    '''
    Node with no children.
    '''
    def __init__(self):
        self._children = None
        super(Leaf, self).__init__()

class ConsumerNode(Leaf):
    def __init__(self):
        super(ConsumerNode, self).__init__()
    
    def consume(self, item):
        '''
        Method definition that needs to be implemented by subclasses. 

        - Arguments:
            - item: the item being received as input (or consumed).
        '''
        raise NotImplementedError('consume function needs to be implemented\
                            by subclass')

class ProcessorNode(Node):
    def __init__(self, nb_tasks : int = 1, device_type = CPU):
        self._nb_tasks = nb_tasks
        if device_type not in device_types:
            raise ValueError('Device is not one of {}'.format(",".join(device_types)))
        self._device_type = device_type
        super(ProcessorNode, self).__init__()

    @property
    def nb_tasks(self):
        '''
        Returns the number of tasks to allocate to this processor
        '''
        return self._nb_tasks
    
    @property
    def device_type(self):
        '''
        Returns the preferred device type to use to run the processor's code
        '''
        return self._device_type
    
    def change_device(self, device_type):
        if device_type not in device_types:
            raise ValueError('Device is not one of {}'.format(",".join(device_types)))
        self._device_type = device_type
    
    def process(self, inp : any) -> any:
        '''
        Method definition that needs to be implemented by subclasses.

        - Arguments:
            - inp: object or list of objects being received for processing \
                from parent nodes.
        
        - Returns:
            - the output being consumed by child nodes.
        '''
        raise NotImplementedError('process function needs to be implemented\
                            by subclass')

class OneTaskProcessorNode(ProcessorNode):
    '''
    Used for processes that keep internal state so they are easily parallelizable.
    The main use of this class if for processes that can only run one
    task, such as trackers and aggregators.
    '''
    def __init__(self):
        super(OneTaskProcessorNode, self).__init__(nb_tasks = 1)

class ExternalProcessorNode(ProcessorNode):
    def __init__(self, processor : Processor, nb_proc : int = 1, device = CPU):
        self._processor = processor
        super(ExternalProcessorNode, self).__init__(nb_proc, device)
    
    def process(self, inp):
        return self._processor.process(inp)

class FunctionProcessorNode(ProcessorNode):
    def __init__(self, processor_function, nb_proc : int = 1, device = CPU):
        self._fn = processor_function
        super(FunctionProcessorNode, self).__init__(nb_proc, device)
    
    def process(self, inp):
        return self._fn(inp)

class ProducerNode(Node):
    '''
    The `producer node` does not receive input, and produces input. \
        Each time the ``next()`` method is called, it produces a new input.
    
    It would have been more natural to implement the ``ProducerNode`` as a generator, \
        but generators cannot be pickled, and hence you cannot easily work with generators \
        in a multiprocessing setting.
    '''
    def __init__(self):
        super(ProducerNode, self).__init__()

    def next(self) -> any:
        '''
        Returns next produced element.

        Raises ``StopIteration`` after the last element has been produced
        and a call to self.next happens.
        '''
        raise NotImplementedError('Method needs to be implemented by subclass')



'''
The ``videoflow.processors.vision`` package contains processors
that are used in the computer vision domain, such as detectors,
classifiers, trackers, pose estimation, etc.
'''

from .annotators import ImageAnnotator, BoundingBoxAnnotator, TrackerAnnotator
from .detectors import ObjectDetector, TensorflowObjectDetector
from .trackers import BoundingBoxTracker, KalmanFilterBoundingBoxTracker



'''
This package contains a collection of consumers that implement the ``videoflow.core.node.ConsumerNode`` \
    interface
'''
from .basic import CommandlineConsumer
from .video import VideofileWriter


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .task import Task
from .node import Node
from multiprocessing import Process

class Messenger:
    '''Not
    Utility class that tasks use to receive input and write output. \
    The ``videoflow.core.task.Task`` class knows what are the graph nodes from \
    where it receives computation outputs, and knows what are the graph nodes \
    that depend on its computation, but is oblivious of how to communicate with them. \
    The messenger, which is tightly coupled with the `execution environment` being used, \
    knows how to do this for the task.
    '''
    def publish_message(self, message):
        '''
        Publishes output message to a place where the child task will receive it.
        Depending on the kind of environment, this method might drop the message
        if the receiving container (usually a queue) is full.
        '''
        raise NotImplementedError('Messenger subclass must implement method')
    
    def passthrough_message(self):
        '''
        Used when the task has received a message that might be needed by another \
        task below it, but when the task itself does not produces output needed by \
        tasks below it. (i.e.: ``videoflow.core.task.ConsumerTask``)

        Depending on the kind of environment, this method might drop the message \
        if the receiving container (usually a queue) is full.
        '''
        raise NotImplementedError('Messenger subclass must implement method')
    
    def passthrough_termination_message(self):
        '''
        Same as ``passthrough_message``, but this method will never drop the message regardless \
        of environment, which means that sometimes this method might block until it can deliver \ 
        the message.
        '''
        raise NotImplementedError('Messenger subclass must implement method')

    def publish_termination_message(self, message):
        '''
        Similar to ``publish_message``, but this method will never drop the message regardless \
        of environment, which means that sometimes this method might block until it can deliver \ 
        the message.
        '''
        raise NotImplementedError('Messenger subclass must implement method')

    def check_for_termination(self) -> bool:
        '''
        Returns true if a flow termination signal has been received.  Used by ``videoflow.core.task.ProducerTask``.
        '''
        raise NotImplementedError('Messenger subclass must implement method')

    def receive_message(self):
        '''
        This method blocks. It waits until a message has been received.

        - Returns:
            - message: the message received from parent task in topological sort.
        '''
        raise NotImplementedError('Messenger subclass must implement method')


class ExecutionEnvironment:
    '''
    Defines the interface of the `execution environment`
    '''
    def __init__(self):
        pass
    
    def _al_create_communication_channels(self, tasks):
        raise NotImplementedError('Subclass of ExecutionEnvironment must implement')
    
    def _al_create_and_set_messengers(self, tasks):
        raise NotImplementedError('Subclass of ExecutionEnvironment must implement')
    
    def _al_create_and_start_processes(self, tasks):
        raise NotImplementedError('Subclass of ExecutionEnvironment must implement')
    
    def signal_flow_termination(self):
        '''
        Signals the execution environment that the flow needs to stop. \
        When this signal is received, all consumer tasks will pick it on \
        and pass it together with the flow until they reach every task in \
        the graph and everyone stops working.
        '''
        raise NotImplementedError('Subclass of ExecutionEnvironment must implement')
    
    def join_task_processes(self):
        '''
        Blocking method.  It is supposed to make the calling process sleep until all task \
        processes have finished processing.
        '''
        raise NotImplementedError('Subclass of ExecutionEnvironment must implement')

    def allocate_and_run_tasks(self, tasks):
        '''
        Defines a template with the order of methods that need to run in order to \
        allocate and run tasks.  How those methods are implemented corresponds to \
        subclasses of this class that implement different `execution environments`.

        '''
        self._al_create_communication_channels(tasks)
        self._al_create_and_set_messengers(tasks)
        self._al_create_and_start_processes(tasks)
    
    

    

    



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from ..core.node import OneTaskProcessorNode

class SumAggregator(OneTaskProcessorNode):
    '''
    Keeps a running sum of all the inputs processed
    '''
    def __init__(self):
        self._sum = 0
        super(SumAggregator, self).__init__()
    
    def process(self, inp):
        '''
        - Arguments:
            - inp: a number
        
        - Returns:
            - sum: the cumulative sum up to this point, including ``inp`` in it.
        '''
        self._sum += inp
        return self._sum

class MultiplicationAggregator(OneTaskProcessorNode):
    '''
    Keeps a running multiplication of all the inputs processed
    '''
    def __init__(self):
        self._mult = 1
        super(MultiplicationAggregator, self).__init__()

    def process(self, inp):
        '''
        - Arguments:
            - inp: a number
        
        - Returns:
            - mult: the cumulative multiplication up to this point, including ``inp`` in it.
        '''
        self._mult *= inp
        return self._mult  

class CountAggregator(OneTaskProcessorNode):
    '''
    Keeps count of all the items processed
    '''
    def __init__(self):
        self._count = 0
        super(CountAggregator, self).__init__()
    
    def process(self, inp):
        '''
        - Arguments:
            - inp: a number
        
        - Returns:
            - count: the cumulative count up to this point
        '''
        self._count += 1
        return self._count

class MaxAggregator(OneTaskProcessorNode):
    def __init__(self):
        self._max = float("-inf")
        super(MaxAggregator, self).__init__()
    
    def process(self, inp):
        '''
        - Arguments:
            - inp: a number
        
        - Returns:
            - max: the maximum seen value up to this point
        '''
        if inp > self._max:
            self._max = inp
        return self._max

class MinAggregator(OneTaskProcessorNode):
    def __init__(self):
        self._min = float("inf")
        super(MinAggregator, self).__init__()
    
    def process(self, inp):
        '''
        - Arguments:
            - inp: a number
        
        - Returns:
            - min: the the minimum seen value up to this point
        '''
        if inp < self._min:
            self._min = inp
        return self._min



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import cv2
import numpy as np

from ..core.node import ConsumerNode

class VideofileWriter(ConsumerNode):
    '''
    Opens a video file writer object and writes subsequent \
    frames received into the object.  If video file exists \
    it overwrites it.
    
    The video writer will open when it receives the first frame.

    - Arguments:
        - video_file: path to video.  Folder where video lives must exist. Extension must be .avi
        - fps: frames per second
    '''
    def __init__(self, video_file : str, fps : int = 30):
        if video_file[-4:] != '.avi':
            raise ValueError("Video extension must be .avi")
        self._video_file = video_file
        self._fps = fps
        self._out = None
        super(VideofileWriter, self).__init__()

    def open(self):
        '''
        We don't open the video writer here.  We need to wait for the first frame \
        to arrive in order to determine the width and height of the video
        '''
        pass
    
    def close(self):
        '''
        Closes the video stream
        '''
        if self._out is not None and self._out.isOpened():
            self._out.release()

    def consume(self, item : np.array):
        '''
        Receives the picture frame to append to the video and appends it to the video.
        
        If it is the first frame received, it opens the video file and determines \
            the height and width of the video from the dimensions of that first frame. \
            Every subsequent frame is expected to have the same height and width. If it \
            does not has it, it gets resized to it.
        
        
        - Arguments:
            - item: np.array of dimension (height, width, 3)
        '''
        if self._out is None:
            self._height = item.shape[0]
            self._width = item.shape[1]
            self._out = cv2.VideoWriter(self._video_file, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G'), self._fps, (self._width, self._height))
        
        resized = cv2.resize(item, (self._width, self._height), interpolation = cv2.INTER_AREA)
        self._out.write(resized)


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
import cv2

from ...core.node import ProcessorNode
from ...utils.parsers import parse_label_map

class ImageAnnotator(ProcessorNode):
    '''
    Interface for all image annotators. 
    All image annotators receive as input an image and annotation
    metadata, and return as output a copy of the image with
    the drawings representing the metadata.
    '''
    def __init__(self, nb_tasks = 1):
        super(ImageAnnotator, self).__init__(nb_tasks = nb_tasks)
    
    def _annotate(self, im : np.array, annotations : any) -> np.array:
        raise NotImplementedError('Subclass must implement this method')

    def process(self, im : np.array, annotations : any) -> np.array:
        '''
        Returns a copy of ``im`` visually annotated with the annotations defined in `annotations`
        '''
        to_annotate = np.array(im)
        return self._annotate(to_annotate, annotations)
        
class BoundingBoxAnnotator(ImageAnnotator):
    '''
    Draws bounding boxes on images.
    - Arguments:
        - class_labels_path: path to pbtxt file that defines the labels indexes
        - box_color: color to use to draw the boxes
        - box_thickness: thickness of boxes to draw
        - text_color: color of text to draw
    '''
    def __init__(self, class_labels_path, box_color = (255, 225, 0), box_thickness = 2, text_color = (255, 255, 0), nb_tasks = 1):
        self._box_color = box_color
        self._text_color = text_color
        self._box_thickness = box_thickness
        self._index_label_d = parse_label_map(class_labels_path)
        super(BoundingBoxAnnotator, self).__init__(nb_tasks = nb_tasks)

    def _annotate(self, im : np.array, boxes : np.array) -> np.array:
        '''
        - Arguments:
            - im: np.array
            - boxes: np.array of shape (nb_boxes, 6) \
                second dimension entries are [ymin, xmin, ymax, xmax, class_index, score]
        
        - Returns:
            - annotated_im: image with the visual annotations embedded in it.
        '''

        for i in range(len(boxes)):
            bbox = boxes[i]
            ymin, xmin, ymax, xmax = int(bbox[0]), int(bbox[1]), int(bbox[2]), int(bbox[3])
            y_label = ymin - 15 if ymin - 15 > 15 else min(ymin + 15, ymax)
            klass_id = int(bbox[4])
            klass_text = self._index_label_d[klass_id]
            confidence = bbox[5]
            label = "{}: {:.2f}%".format(klass_text, confidence * 100)
            cv2.rectangle(im, (xmin, ymin), (xmax, ymax), self._box_color, self._box_thickness)
            cv2.putText(im, label, (xmin, y_label), cv2.FONT_HERSHEY_SIMPLEX, 0.5, self._text_color, lineType = cv2.LINE_AA)
        return im

class TrackerAnnotator(ImageAnnotator):
    '''
    Draws bounding boxes on images with track id.
    '''
    def __init__(self, box_color = (255, 225, 0), box_thickness = 2, text_color = (255, 255, 255), nb_tasks = 1):
        self._box_color = box_color
        self._text_color = text_color
        self._box_thickness = box_thickness
        super(TrackerAnnotator, self).__init__(nb_tasks = nb_tasks)

    def _annotate(self, im : np.array, boxes : np.array) -> np.array:
        '''
        - Arguments:
            - im: np.array
            - boxes: np.array of shape (nb_boxes, 5) \
                second dimension entries are [ymin, xmin, ymax, xmax, track_id]
        
        - Returns:
            - annotated_im: image with the visual annotations embedded in it.
        '''
        for i in range(len(boxes)):
            bbox = boxes[i]
            ymin, xmin, ymax, xmax = int(bbox[0]), int(bbox[1]), int(bbox[2]), int(bbox[3])
            y_label = ymin - 15 if ymin - 15 > 15 else min(ymin + 15, ymax)
            track_id = bbox[4]
            label = "{}".format(track_id)
            cv2.rectangle(im, (xmin, ymin), (xmax, ymax), self._box_color, self._box_thickness)
            cv2.putText(im, label, (xmin, y_label), cv2.FONT_HERSHEY_SIMPLEX, 0.5, self._text_color, lineType = cv2.LINE_AA)
        return im





from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
from multiprocessing import Process, Queue, Event

from ..core.node import Node, GPU, CPU
from ..core.task import Task, ProducerTask, ProcessorTask, ConsumerTask
from ..core.environment import ExecutionEnvironment, Messenger
from ..utils.system import get_number_of_gpus

def task_executor_fn(task : Task):
    task.run()

def task_executor_gpu_fn(task : Task, gpu_id : int):
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)
    task.run()

def allocate_process_task(task):
    proc = Process(target = task_executor_fn, args = (task,))
    return proc

def allocate_process_task_gpu(task, gpu_id):
    proc = Process(target = task_executor_gpu_fn, args = (task, gpu_id))
    return proc

class BatchprocessingQueueMessenger(Messenger):
    '''
    BatchprocessingQueueMessenger is a messenger that communicates
    through queues of type ``multiprocessing.Queue``.  It is not real
    time, which means that if a queue is full when publishing a 
    message to it, it will block until the queue can process it.
    '''
    def __init__(self, computation_node : Node, task_queue : Queue, parent_task_queue : Queue,
        termination_event : Event):
        self._computation_node = computation_node
        self._parent_task_queue = parent_task_queue
        self._task_queue = task_queue
        self._parent_nodes_ids = []
        if self._computation_node.parents is not None:
            self._parent_nodes_ids = [a.id for a in self._computation_node.parents]
        self._termination_event = termination_event
        self._last_message_received = None
    
    def publish_message(self, message):
        '''
        Publishes output message to a place where the child task will receive it. \
        Will drop the message is the receiving queue is full.
        '''
        if self._last_message_received is None:
            msg = {
                self._computation_node.id : message
            }
            self._task_queue.put(msg, block = True)
        else:
            self._last_message_received[self._computation_node.id] = message
            self._task_queue.put(self._last_message_received, block = True)
    
    def check_for_termination(self) -> bool:
        '''
        Checks if someone has set a termination event.
        '''
        return self._termination_event.is_set()

    def publish_termination_message(self, message):
        '''
        This method is identical to publish message
        '''
        return self.publish_message(message)

    def passthrough_message(self):
        self._task_queue.put(self._last_message_received, block = True)
    
    def passthrough_termination_message(self):
        return self.passthrough_message()
    
    def receive_message(self):
        input_message_dict = self._parent_task_queue.get()
        self._last_message_received = input_message_dict
        inputs = [input_message_dict[a] for a in self._parent_nodes_ids]
        return inputs
    
class RealtimeQueueMessenger(Messenger):
    '''
    RealtimeQueueMessenger is a messenquer that communicates through
    queues of type ``multiprocessing.Queue``.  It is a real time, which 
    means that if a queue is full when publishing a message to it,
    it will drop the message and not block.  The methods that 
    publish and passthrough termination messages will block and not drop.
    '''
    def __init__(self, computation_node : Node, task_queue : Queue, parent_task_queue : Queue,
                termination_event : Event):
        self._computation_node = computation_node
        self._parent_task_queue = parent_task_queue
        self._task_queue = task_queue
        self._parent_nodes_ids = []
        if self._computation_node.parents is not None:
            self._parent_nodes_ids = [a.id for a in self._computation_node.parents]
        self._termination_event = termination_event
        self._last_message_received = None

    def publish_message(self, message):
        '''
        Publishes output message to a place where the child task will receive it. \
        Will drop the message is the receiving queue is full.
        '''
        if self._last_message_received is None:
            try:
                msg = {
                    self._computation_node.id : message
                }
                self._task_queue.put(msg, block = False)
            except:
                pass
        else:
            self._last_message_received[self._computation_node.id] = message
            try:
                self._task_queue.put(self._last_message_received, block = False)
            except:
                pass
    
    def check_for_termination(self) -> bool:
        '''
        Checks if someone has set a termination event.
        '''
        return self._termination_event.is_set()

    def publish_termination_message(self, message):
        '''
        This method is identical to publish message, but is blocking
        Because, the termination message cannot be dropped.
        '''
        if self._last_message_received is None:
            try:
                msg = {
                    self._computation_node.id : message
                }
                self._task_queue.put(msg, block = True)
            except:
                pass
        else:
            self._last_message_received[self._computation_node.id] = message
            try:
                self._task_queue.put(self._last_message_received, block = True)
            except:
                pass

    def passthrough_message(self):
        try:
            self._task_queue.put(self._last_message_received, block = False)
        except:
            pass
    
    def passthrough_termination_message(self):
        try:
            self._task_queue.put(self._last_message_received, block = True)
        except:
            pass

    def receive_message(self):
        input_message_dict = self._parent_task_queue.get()
        self._last_message_received = input_message_dict
        inputs = [input_message_dict[a] for a in self._parent_nodes_ids]
        return inputs

class QueueExecutionEnvironment(ExecutionEnvironment):
    def __init__(self):
        self._procs = []
        self._tasks = []
        self._task_output_queues = {}
        self._task_termination_notification_queues = {}
        self._termination_event = None
        self._nb_available_gpus = get_number_of_gpus()
        self._next_gpu = -1
        super(QueueExecutionEnvironment, self).__init__()
    
    def _al_create_communication_channels(self, tasks):
        #1. Create output queues
        for task in tasks:
            queue = Queue(10)
            self._task_output_queues[task.id] = queue
        
        self._termination_event = Event()

    def _al_create_and_start_processes(self, tasks):
        #2. Create processes
        for task in tasks:
            if isinstance(task, ProcessorTask):
                if task.device_type == GPU:
                    self._next_gpu += 1
                    if self._next_gpu < self._nb_available_gpus:
                        proc = allocate_process_task_gpu(task, self._next_gpu)
                    else:
                        try:
                            task.change_device(CPU)
                            proc = allocate_process_task(task)
                        except:
                            raise RuntimeError('No GPU available to allocate {}'.format(str(task._computation_node)))
                else:
                    proc = allocate_process_task(task)
            else:
                proc = allocate_process_task(task)
            self._procs.append(proc)
        
        #3. Start processes
        for proc in self._procs:
            proc.start()
    
    def signal_flow_termination(self):
        self._termination_event.set()
    
    def join_task_processes(self):
        for proc in self._procs:
            proc.join()

class BatchprocessingQueueExecutionEnvironment(QueueExecutionEnvironment):
    def _al_create_and_set_messengers(self, tasks):
        for task in tasks:
            task_queue = self._task_output_queues.get(task.id)
            parent_task_queue = self._task_output_queues.get(task.parent_id, None)
            computation_node = task.computation_node
            messenger = BatchprocessingQueueMessenger(computation_node, task_queue, parent_task_queue, self._termination_event)
            task.set_messenger(messenger)    

class RealtimeQueueExecutionEnvironment(QueueExecutionEnvironment):
    def _al_create_and_set_messengers(self, tasks):
        for task in tasks:
            task_queue = self._task_output_queues.get(task.id)
            parent_task_queue = self._task_output_queues.get(task.parent_id, None)
            computation_node = task.computation_node
            messenger = RealtimeQueueMessenger(computation_node, task_queue, parent_task_queue, self._termination_event)
            task.set_messenger(messenger)


"""Setup script for realpython-reader"""

import os.path
import setuptools
from setuptools import setup

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
exec(open('videoflow/version.py').read())
setup(
    name = "videoflow",
    version = __version__,
    description="Python video streams processing library",
    long_description = README,
    long_description_content_type = "text/markdown",
    url = "https://github.com/jadielam/videoflow",
    author = "Jadiel de Armas",
    author_email = "jadielam@gmail.com",
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages = setuptools.find_packages(),
    include_package_data = True,
    install_requires = [
        "opencv-python",
        "filterpy",
        "scikit-learn"
    ]
)


from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import tensorflow as tf

class TensorflowModel:
    def __init__(self, pb_file_path, input_tensors_names, output_tensors_names, device_id = "cpu:0"):
        '''
        - Arguments:
            - pb_file_path (str): path to pb file
            - input_tensors_names (list(str)): list of names of input tensors
            - output_tensors_names (list(str)): list of names of output tensors
            - device_id (str): name of device where model should be allocated. Model allocation \
                in that device is not guaranteed. If the device cannot be found, it will \
                default to cpu.
        '''
        self._pb_file_path = pb_file_path
        self._device_id = device_id
        self._output_tensors_names = output_tensors_names
        self._input_tensors_names = input_tensors_names
        self._session = None
        self._model_graph = None
        self._output_tensors = None
        self._input_tensors = None
        self._load_model()
    
    def _load_model(self):
        '''
        Loads model from file and creates the model session to make it ready
        for inference.
        '''
        with tf.device(self._device_id):
            self._model_graph = tf.Graph()
            with self._model_graph.as_default():
                graph_def = tf.GraphDef()
                with tf.gfile.GFile(self._pb_file_path, 'rb') as fid:
                    serialized_graph = fid.read()
                    graph_def.ParseFromString(serialized_graph)
                    tf.import_graph_def(graph_def, name = '')
        
        self._session = tf.Session(graph = self._model_graph)
        self._output_tensors = [self._model_graph.get_tensor_by_name(name) for name in self._output_tensors_names]
        self._input_tensors = [self._model_graph.get_tensor_by_name(name) for name in self._input_tensors_names]

    def _close_session(self):
        '''
        Closes the tensorflow session that was opened when the model was loaded.
        '''
        self._session.close()
    
    def run_on_input(self, *inp_l):
        '''
        - Arguments:
            - inp_l: a list of inputs to be passed to the model. \
                Must be given in the same order that the list of `input_tensors_names` was given to the constructor.

        - Returns:
            - output_l: a list of outputs of the same length and in the same order as the `output_tensors_names` \
                was provided to the constructor.
        '''
        if self._session is None:
            self._load_model()
        feed_dict = dict(zip(self._input_tensors, inp_l))
        output_l = self._session.run(
            self._output_tensors,
            feed_dict = feed_dict
        )
        return output_l



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numpy as np
import zmq

REFERENCE_SOCKET_NUMBER = 5000

def termination_socket_from_task_id(task_id : int):
    return REFERENCE_SOCKET_NUMBER - task_id

def socket_from_task_id(task_id : int):
    return REFERENCE_SOCKET_NUMBER + task_id

def task_id_from_socket(socket_number : int):
    return socket_number - REFERENCE_SOCKET_NUMBER

def task_id_from_termination_socket(socket_number : int):
    return REFERENCE_SOCKET_NUMBER - socket_number

def recv_zipped_pickle(socket: zmq.Socket, flags: int=0):
    """
    Receive a sent zipped pickle.
    """
    message = socket.recv(flags)
    object = zlib.decompress(message)
    return pickle.loads(object)

def send_zipped_pickle(socket: zmq.Socket, obj: any, flags: int=0, protocol: int=-1):
    """
    Pickle an object, and zip the pickle before sending it
    """
    object = pickle.dumps(obj, protocol)
    compressed_object = zlib.compress(object)
    return socket.send(compressed_object, flags=flags)


def send_array(socket: zmq.Socket, array: np.array, flags: int=0, copy: bool=True, track: bool=False):
    """
    Send a numpy array with metadata, type and shape
    """
    dictionary = dict(
        dtype = str(array.dtype),
        shape = array.shape,
    )
    socket.send_json(dictionary, flags|zmq.SNDMORE)
    return socket.send(array, flags, copy=copy, track=track)


def recv_array(socket: zmq.Socket, flags: int=0, copy: int=True, track: bool=False):
    """
    Receive a numpy array
    """
    dictionary = socket.recv_json(flags=flags)
    message = socket.recv(flags=flags, copy=copy, track=track)
    buffer = memoryview(message)
    array = np.frombuffer(buffer, dtype=dictionary['dtype'])
    return array.reshape(dictionary['shape'])